from dataclasses import dataclass
from typing import Optional

from lkb.lkb_base_class import Concept
from .basic_lkb import Basic_Lexical_Knowledge_Base
from .neo4j_lkb import Neo4J_Lexical_Knowledge_Base
from .rdf_lkb import RDF_Lexical_Knowledge_Base
from .wordnet_lkb import WordNet_Lexical_Knowledge_Base

from fileutils import cached_path
import os

from dotenv import load_dotenv


def URI():
    return os.environ["NEO4J_URI"]


@dataclass(frozen=True)
class TRENDNet_Concept(Concept):
    concept_name: str
    semantic_type: str
    definition: Optional[str] = "NO DEFINITION"


class LKBInstanceDefinition:
    def __init__(self, data_source_path, concept_class):
        self._concept_class = concept_class
        self._data_source_path = data_source_path

    @property
    def concept_class(self):
        return self._concept_class

    @property
    def data_source_path(self):
        return cached_path(self._data_source_path)


LKBS = {
    "TRENDNET-0.1.0": LKBInstanceDefinition(
        "https://databrickstrenddata/lkb/trendnet.json",
        TRENDNet_Concept,
    ),
    "TRENDNET-0.2.0": LKBInstanceDefinition(
        "https://databrickstrenddata/lkb/trendnet-0.2.0.json",
        TRENDNet_Concept,
    ),
    "TRENDNET-0.2.1": LKBInstanceDefinition(
        "https://databrickstrenddata/lkb/trendnet_0.2.1.json",
        TRENDNet_Concept,
    ),
}

DAOS = {
    "BASIC": Basic_Lexical_Knowledge_Base,
    "NEO4J": Neo4J_Lexical_Knowledge_Base,
    "RDF": RDF_Lexical_Knowledge_Base,
    "WORDNET_LIKE": WordNet_Lexical_Knowledge_Base,
}


def load_lkb(lkb_name, dao="BASIC"):
    instance_definition = LKBS[lkb_name]
    data_path = instance_definition.data_source_path
    concept_class = instance_definition.concept_class
    if dao in DAOS:
        dao = DAOS[dao]
    else:
        raise ValueError(f"No such DAO:{dao}")
    return dao.read_json(data_path, concept_class)


def lkb_connect(lkb_name, uri=None, dao="NEO4J"):
    if uri is None:
        uri = URI()
    if dao != "NEO4J":
        raise ValueError("Can only connect to a Neo4j LKB (for now).")
    else:
        instance_definition = LKBS[lkb_name]
        dao = DAOS[dao]
        return dao.connect(instance_definition.concept_class, uri=uri)
