from __future__ import annotations
from abc import abstractmethod
import json
from dataclasses import asdict, dataclass


@dataclass(frozen=True)
class Concept:
    id: str


@dataclass(frozen=True)
class Term:
    id: str
    string: str


@dataclass(frozen=True)
class Conceptual_Edge:
    concept_id_1: str
    concept_id_2: str
    rel_id: str


@dataclass(frozen=True)
class Lexical_Edge:
    concept_id: str
    term_id: str


@dataclass(frozen=True)
class Conceptual_Relation:
    id: str
    string: str  # relation name


class Lexical_Knowledge_Base:
    def __init__(self, concept_type):
        self._concept_type = concept_type

    @classmethod
    def import_lkb(
        cls, concepts, terms, conceptual_relations, lexical_edges, conceptual_edges
    ):
        concept_type = concepts[0].__class__
        lkb = cls(concept_type)
        lkb.post_init(
            concepts, terms, conceptual_relations, lexical_edges, conceptual_edges
        )
        return lkb

    @abstractmethod
    def post_init(
        self, concepts, terms, conceptual_relations, lexical_edges, conceptual_edges
    ):
        raise NotImplementedError

    @abstractmethod
    def get_concept(self, concept_id: str):
        pass

    @abstractmethod
    def get_term(self, term_id):
        pass

    @abstractmethod
    def get_relation(self, rel_id):
        pass

    @abstractmethod
    def get_outward_edges(self, subject_concept_id: str):
        pass

    @abstractmethod
    def get_inward_edges(self, object_concept_id: str):
        pass

    @abstractmethod
    def get_terms_from_concept_id(self, concept_id: str):
        pass

    @abstractmethod
    def get_concepts_from_term_id(self, term_id: str):
        pass

    @abstractmethod
    def get_data(self):
        pass

    def get_named_concepts(self):
        # return a list of concepts, and a string corresponding to their canonical term.
        raise NotImplementedError

    @classmethod
    def read_json(cls, file_path, concrete_concept_class=Concept):
        with open(file_path, "r") as infile:
            dictionary = json.load(infile)
            concepts = [
                concrete_concept_class(**concept) for concept in dictionary["concepts"]
            ]
            terms = [Term(**term) for term in dictionary["terms"]]
            conceptual_edges = [
                Conceptual_Edge(**conceptual_edge)
                for conceptual_edge in dictionary["conceptual_edges"]
            ]
            lexical_edges = [
                Lexical_Edge(**lexical_edge)
                for lexical_edge in dictionary["lexical_edges"]
            ]
            conceptual_relations = [
                Conceptual_Relation(**conceptual_relation)
                for conceptual_relation in dictionary["conceptual_relations"]
            ]
            return cls.import_lkb(
                concepts,
                terms,
                conceptual_relations,
                lexical_edges,
                conceptual_edges,
            )

    def write_json(self, file_path):
        (
            concepts,
            terms,
            conceptual_edges,
            lexical_edges,
            conceptual_relations,
        ) = self.get_data()
        dictionary = {
            "concepts": list(concepts),
            "terms": list(terms),
            "conceptual_edges": list(conceptual_edges),
            "lexical_edges": list(lexical_edges),
            "conceptual_relations": list(conceptual_relations),
        }
        jsonified = {
            key: [asdict(entity) for entity in value]
            for key, value in dictionary.items()
        }
        with open(file_path, "w") as outfile:
            json.dump(jsonified, outfile, indent=2)
