import csv
from dataclasses import asdict
import os
from pathlib import Path

from dotenv import load_dotenv
from lkb.lkb_base_class import (
    Conceptual_Relation,
    Lexical_Knowledge_Base,
    Term,
)
from neo4j import GraphDatabase
import neo4j

AUTH = None


def getNeo4jImportPath():
    return os.environ["NEO4J_IMPORT_PATH"]
    # else:
    #     # TODO: Use docker to run lkb, then we can mount volumes
    #     #       from neo4j into that container and set
    #     #       NEO4J_IMPORT_PATH directly and remove code below.
    #     source_path = Path(__file__).resolve()
    #     return os.path.join(source_path.parent, "..", "..", "neo4j", "import")


class Neo4J_Lexical_Knowledge_Base(Lexical_Knowledge_Base):
    def post_init(
        self,
        concepts,
        terms,
        conceptual_relations,
        lexical_edges,
        conceptual_edges,
    ):
        self.driver = GraphDatabase.driver(self.uri(), auth=AUTH)
        self.csv_directory = getNeo4jImportPath()
        self.write_concepts(concepts)
        self.write_terms(terms)
        self.write_conceptual_relations(conceptual_relations, conceptual_edges)
        self.write_lexical_edges(lexical_edges)
        print("Done loading into Neo4j.")

    @staticmethod
    def uri():

        return os.environ["NEO4J_URI"]

    @classmethod
    def delete_db(cls):
        driver = GraphDatabase.driver(cls.uri())

        def delete_data(tx):
            tx.run("MATCH (n) DETACH DELETE n")

        def check_for_indices(tx):
            return tx.run("CALL db.indexes").peek() is not None

        def delete_indices(tx):
            tx.run("DROP INDEX term_index")
            tx.run("DROP INDEX concept_index")

        with driver.session() as session:
            session.write_transaction(delete_data)
            print(session.read_transaction(check_for_indices))
            contains_indices = session.read_transaction(check_for_indices)
            print(contains_indices)
            if contains_indices:
                session.write_transaction(delete_indices)
            # session.write_transaction(delete_indices)

    def write_terms(self, terms):
        print("Adding terms to CSV...")
        # add terms
        terms_filename = os.path.join(self.csv_directory, "terms.csv")
        with open(terms_filename, "w", newline="") as f:
            writer = csv.writer(f)
            attrs = list(terms[0].__annotations__.keys())
            header = ["Entity"] + attrs
            writer.writerow(header)
            for term in terms:
                term_dict = asdict(term)
                writer.writerow(
                    ["Term"] + [term_dict[attr_name] for attr_name in attrs]
                )
        term_attr_string = ",".join(
            [f"{attr_name} : row.{attr_name}" for attr_name in attrs]
        )
        print("Adding terms to Neo4j...")
        query = """
                USING PERIODIC COMMIT
                LOAD CSV WITH HEADERS FROM '{0}' AS row
                CREATE (t:Term {{{1}}})
            """.format(
            "file:///terms.csv", term_attr_string
        )
        with self.driver.session(default_access_mode=neo4j.WRITE_ACCESS) as session:
            session.run(query)
            session.run("CREATE INDEX term_index FOR (t:Term) ON (t.id)").consume()

    def write_concepts(self, concepts):
        print("Adding concepts to CSV...")
        concepts_filename = os.path.join(self.csv_directory, "concepts.csv")
        with open(concepts_filename, "w", newline="") as f:
            writer = csv.writer(f)
            attrs = set(concepts[0].__annotations__.keys())
            attrs.add("id")
            attrs = list(attrs)
            header = ["Entity"] + attrs
            writer.writerow(header)
            for concept in concepts:
                concept_dict = asdict(concept)
                writer.writerow(
                    ["Concept"] + [concept_dict[attr_name] for attr_name in attrs]
                )
        concept_attr_string = ", ".join(
            [f"{attr_name} : row.{attr_name}" for attr_name in attrs]
        )
        print("Adding concepts to Neo4j...")
        query = """
                USING PERIODIC COMMIT
                LOAD CSV WITH HEADERS FROM '{0}' AS row
                CREATE (c:Concept {{{1}}})
            """.format(
            "file:///concepts.csv", concept_attr_string
        )
        with self.driver.session(default_access_mode=neo4j.WRITE_ACCESS) as session:
            session.run(query)
            session.run(
                "CREATE INDEX concept_index FOR (c:Concept) ON (c.id)"
            ).consume()

    def write_conceptual_relations(self, conceptual_relations, conceptual_edges):
        self._conceptual_relations = {
            relation.id: relation for relation in conceptual_relations
        }
        conceptual_edges_filename = os.path.join(
            self.csv_directory, "conceptual_edges.csv"
        )
        print("Adding conceptual edges to CSV...")
        with open(conceptual_edges_filename, "w", newline="") as f:
            writer = csv.writer(f)
            header = ["Entity"] + [
                "concept_id_1",
                "concept_id_2",
                "rel_id",
                "rel_string",
            ]
            writer.writerow(header)
            for conceptual_edge in conceptual_edges:
                rel_string = self._conceptual_relations[conceptual_edge.rel_id].string
                writer.writerow(
                    [
                        "Conceptual Edge",
                        conceptual_edge.concept_id_1,
                        conceptual_edge.concept_id_2,
                        conceptual_edge.rel_id,
                        rel_string,
                    ]
                )
        print("Adding conceptual edges to Neo4j...")
        query = """
                USING PERIODIC COMMIT
                LOAD CSV WITH HEADERS FROM '{0}' AS row
                MATCH (c1:Concept {{id: row.concept_id_1}}),(c2:Concept {{id: row.concept_id_2}})
                MERGE (c1)-[r:RELATED_TO]->(c2)
                ON CREATE
                SET r.rel_id = row.rel_id
                SET r.relation = row.rel_string
            """.format(
            "file:///conceptual_edges.csv"
        )
        with self.driver.session(default_access_mode=neo4j.WRITE_ACCESS) as session:
            session.run(query)

    def write_lexical_edges(self, lexical_edges):
        print("Adding lexical edges to CSV...")
        lexical_edges_filename = os.path.join(self.csv_directory, "lexical_edges.csv")
        with open(lexical_edges_filename, "w", newline="") as f:
            writer = csv.writer(f)
            header = ["Entity"] + ["concept_id", "term_id"]
            writer.writerow(header)
            for lexical_edge in lexical_edges:
                writer.writerow(
                    [
                        "Lexical Edge",
                        lexical_edge.concept_id,
                        lexical_edge.term_id,
                    ]
                )
        print("Adding lexical edges to Neo4j...")
        query = """
                USING PERIODIC COMMIT 500
                LOAD CSV WITH HEADERS FROM '{0}' AS row
                MATCH (c:Concept {{id: row.concept_id}}),(t:Term {{id: row.term_id}})
                MERGE (c)-[r:HAS_TERM]->(t)
            """.format(
            "file:///lexical_edges.csv"
        )
        with self.driver.session(default_access_mode=neo4j.WRITE_ACCESS) as session:
            session.run(query)

    def get_concept(self, concept_id: str):
        def _get_concept(tx, concept_id):
            first = tx.run(
                "MATCH (c:Concept {id: $concept_id}) return c",
                concept_id=concept_id,
            ).single()
            return first.value("c")

        with self.driver.session() as session:
            return self._concept_type(
                **dict(session.read_transaction(_get_concept, concept_id))
            )

    def get_term(self, term_id):
        def _get_term(tx, term_id):
            first = tx.run(
                "MATCH (t:Term {id: $term_id}) return t", term_id=term_id
            ).single()
            return first.value("t")

        with self.driver.session() as session:
            return Term(**dict(session.read_transaction(_get_term, term_id)))

    def get_relation(self, rel_id):
        return self._conceptual_relations[rel_id]

    def get_outward_edges(self, subject_concept_id: str):
        def _get_outward_edges(tx, subject_concept_id):
            result = tx.run(
                "MATCH (c1:Concept {id: $subject_concept_id}) -[r]->(c2:Concept) return c2,r",
                subject_concept_id=subject_concept_id,
            )
            return [(row.value("r"), row.value("c2")) for row in result]

        with self.driver.session() as session:
            return [
                (
                    # self._conceptual_relations[edge[0]["rel_id"]],
                    Conceptual_Relation(edge[0]["rel_id"], edge[0]["relation"]),
                    self._concept_type(**dict(edge[1])),
                )
                for edge in session.read_transaction(
                    _get_outward_edges, subject_concept_id
                )
            ]

    def get_inward_edges(self, object_concept_id: str):
        def _get_inward_edges(tx, object_concept_id):
            result = tx.run(
                "MATCH (c1:Concept) -[r]->(c2:Concept {id: $object_concept_id}) return c1,r",
                object_concept_id=object_concept_id,
            )
            return [(row.value("r"), row.value("c1")) for row in result]

        with self.driver.session() as session:
            return [
                (
                    Conceptual_Relation(edge[0]["rel_id"], edge[0]["relation"]),
                    self._concept_type(**dict(edge[1])),
                )
                for edge in session.read_transaction(
                    _get_inward_edges, object_concept_id
                )
            ]

    def get_terms_from_concept_id(self, concept_id: str):
        def _get_terms_from_concept_id(tx, concept_id):
            result = tx.run(
                "MATCH (c:Concept {id:$concept_id})-[r:HAS_TERM]->(t) return t",
                concept_id=concept_id,
            )
            return [row.value("t") for row in result]

        with self.driver.session() as session:
            return [
                Term(**dict(term))
                for term in session.read_transaction(
                    _get_terms_from_concept_id, concept_id
                )
            ]

    def get_concepts_from_term_id(self, term_id: str):
        def _get_concepts_from_term_id(tx, term_id):
            result = tx.run(
                "MATCH (c)-[r:HAS_TERM]->(t:Term {id:$term_id}) return c",
                term_id=term_id,
            )
            return [row.value("c") for row in result]

        with self.driver.session() as session:
            return [
                self._concept_type(**dict(concept))
                for concept in session.read_transaction(
                    _get_concepts_from_term_id, term_id
                )
            ]

    def get_data(self):
        pass

    def _get_and_set_relations(self):
        def _get_relations_query(tx):
            result = tx.run(
                "MATCH (c1:Concept)-[r]->(c2:Concept) RETURN DISTINCT r.rel_id,r.relation"
            )
            return [(row.value("r.rel_id"), row.value("r.relation")) for row in result]

        with self.driver.session() as session:
            self._conceptual_relations = {
                r[0]: Conceptual_Relation(r[0], r[1])
                for r in session.read_transaction(_get_relations_query)
            }

    @classmethod
    def connect(cls, concept_type, uri=None):
        lkb = cls(concept_type)
        lkb._set_driver(uri)
        lkb._get_and_set_relations()
        return lkb

    def _set_driver(self, uri):
        if uri is None:
            uri = self.uri()
        self.driver = GraphDatabase.driver(uri)

    def get_named_concepts(self):
        # return a list of concepts, and a string corresponding to their canonical term.
        raise NotImplementedError
